#!/bin/bash
# YES, THIS NEEDS BASH, NOT /bin/sh (e.g. for <<<).

# Copyright (2013) Jann Horn <jann@thejh.net>

# -f                  for files with weird names
# -u                  for coding mistakes (or against, to be precise)
# -e and -o pipefail  so that we don't have to spam the code with error handling
set -f -u -e -o pipefail

# flags for the build - adjust for your needs
# delete all the generated stuff afterwards (with `rm -r gen`)
CC='gcc'
# -fstack-check isn't just hardening - we do unbounded stack allocations in TPRINTF!
CFLAGS='-O3 -Wall -Werror -Wno-error=strict-aliasing -fPIC -std=c99 -march=native -fstack-check -fstack-protector-all -D_FORTIFY_SOURCE=2'

# create build environment if it doesn't exist yet
mkdir -p gen # contains all generated files
mkdir -p gen/realc # source files, with our preprocessing applied
mkdir -p gen/realc_preprocessed # source files, preprocessed by the C compiler
mkdir -p gen/chash # hashes of the preprocessed C files used to generate files in gen/obj
mkdir -p gen/obj # object files

echo "welcome. your friendly compiler will be \"$CC\" today." >&2
echo "going ahead with CFLAGS=\"$CFLAGS\"..." >&2

# generate header
set +f
# needs correct order, so list them here
cat header.h bufio.h > gen/jh.h
set -f
for source_file in $(ls|grep '\.c$'); do
  echo "extracting header data from $source_file..." >&2
  source_name="$(sed 's|\.c$||' <<< "$source_file")"
  echo "/* ----========   $source_name   ========---- */"
  
  cat "$source_file" |
  sed 's|^PUBLIC_FN \(.*\){|KEEPLINE \1;|g' |
  sed 's|^PUBLIC_CONST |KEEPLINE #define |g' |
  sed 's|^HEADER |KEEPLINE |g' |
  (set +e +o pipefail; grep '^KEEPLINE'; exit 0) |
  sed 's|^KEEPLINE ||g'
  
  echo ''
  echo ''
done >> gen/jh.h

# preprocess all source files
for source_file in $(ls|grep '\.c$'); do
  echo "handling source file $source_file..." >&2
  source_name="$(sed 's|\.c$||' <<< "$source_file")"
  
  # do our own preprocessing
  echo '#include "../jh.h"' > "gen/realc/$source_name.c"
  
  cat "$source_file" |
  (set +e +o pipefail; grep -v '^PUBLIC_CONST '; exit 0) |
  sed 's|^PUBLIC_FN ||g' |
  (set +e +o pipefail; grep -v '^HEADER '; exit 0) |
  sed 's| *JH_ATTR_[A-Z_]*||g' |
  cat >> "gen/realc/$source_name.c"
  if [ $? -ne 0 ]; then exit 1; fi
  
  # do the normal C preprocessing
  echo 'preprocessing...' >&2
  $CC -E "gen/realc/$source_name.c" > "gen/realc_preprocessed/$source_name.i"
  
  # compile if there have been changes since the last time
  echo -n 'checking sha checksum... ' >&2
  file_hash="$(shasum "gen/realc_preprocessed/$source_name.i" | cut -d' ' -f1)"
  # drop errors: that file might well be missing
  set +e
  last_file_hash="$(cat "gen/chash/$source_name" 2>/dev/null)"
  set -e
  if [ "$file_hash" = "$last_file_hash" ]; then
    echo 'unchanged, will not compile again' >&2
  else
    echo 'changed – recompiling.' >&2
    $CC $CFLAGS -c -o "gen/obj/$source_name.o" "gen/realc_preprocessed/$source_name.i"
    echo "$file_hash" > "gen/chash/$source_name"
  fi
done

# ... and link!
cd gen/obj
$CC -shared -Wl,-soname,libjh.so -o ../libjh.so $(ls)
ar rcs ../libjh.a $(ls)
cd ../..
